﻿using ELTE.TicTacToeGame.Persistence;
using System;
using System.Linq;

namespace ELTE.TicTacToeGame.Model
{
    /// <summary>
    /// Tic-Tac-Toe játék modell típusa.
    /// </summary>
    public class TicTacToeModel
    {
        #region Private fields
        
        private Player[,] gameBoard; // játéktábla
        private TextFilePersistence persistence; // adatelérés

        #endregion

        #region Events

        /// <summary>
        /// Játék megnyerésének eseménye.
        /// </summary>
        public event EventHandler<GameWonEventArgs> GameWon;

        /// <summary>
        /// Játék végének eseménye.
        /// </summary>
        public event EventHandler GameOver;

        /// <summary>
        /// Mezőváltozás eseménye.
        /// </summary>
        public event EventHandler<FieldChangedEventArgs> FieldChanged;


        #endregion

        #region Constructors
        
        /// <summary>
        /// Tic-Tac-Toe játék modell példányosítása.
        /// </summary>
        public TicTacToeModel() 
        {
            this.gameBoard = new Player[3, 3]; // mátrix létrehozása
            this.persistence = new TextFilePersistence();

            this.NewGame();
        }

        #endregion

        #region Public properties

        /// <summary>
        /// Lépésszám lekérdezése.
        /// </summary>
        public Int32 StepNumber { get; private set; }

        /// <summary>
        /// Aktuális játékos lekérdezése.
        /// </summary>
        public Player CurrentPlayer { get; private set; }

        /// <summary>
        /// Mezőérték lekérdezése, vagy beállítása.
        /// </summary>
        /// <param name="x">Oszlop index.</param>
        /// <param name="y">Sor index.</param>
        /// <returns>A mező értéke.</returns>
        public Player this[Int32 x, Int32 y]
        {
            get
            {
                if (x < 0 || x > this.gameBoard.GetLength(0)) // ellenőrizzük a tartományt
                    throw new ArgumentException("Bad column index.", "x");
                if (y < 0 || y > this.gameBoard.GetLength(1))
                    throw new ArgumentException("Bad row index.", "y");

                return this.gameBoard[x, y];
            }
        }

        #endregion

        #region Public methods

        /// <summary>
        /// Új játék indítása.
        /// </summary>
        public void NewGame() 
        { 
            for (Int32 i = 0; i < this.gameBoard.GetLength(0); i++) // végigmegyünk a mátrix elemein
                for (Int32 j = 0; j < this.gameBoard.GetLength(1); j++)
                {
                    this.gameBoard[i, j] = Player.NoPlayer; // a játékosok pozícióit töröljük
                }

            this.StepNumber = 0;
            this.CurrentPlayer = Player.PlayerX; // először az X lép
        }

        /// <summary>
        /// Játék léptetése.
        /// </summary>
        /// <param name="x">Oszlop index.</param>
        /// <param name="y">Sor index.</param>
        public void StepGame(Int32 x, Int32 y) 
        {
            if (x < 0 || x > this.gameBoard.GetLength(0)) // ellenőrizzük a tartományt
                throw new ArgumentOutOfRangeException("x", "Bad column index.");
            if (y < 0 || y > this.gameBoard.GetLength(1))
                throw new ArgumentOutOfRangeException("y", "Bad row index.");
            if (this.StepNumber >= 9) // ellenőrizzük a lépésszámot
                throw new InvalidOperationException("Game is over!");
            if (this.gameBoard[x, y] != Player.NoPlayer) // ellenőrizzük a mezőt
                throw new InvalidOperationException("Field is not empty!");

            this.gameBoard[x, y] = this.CurrentPlayer; // pozíció rögzítése
            this.OnFieldChanged(x, y, this.CurrentPlayer); // jelezzük egy eseménykiváltással, hogy változott a mező

            this.StepNumber++;
            this.CurrentPlayer = this.CurrentPlayer == Player.PlayerO ? Player.PlayerX : Player.PlayerO;
            // beállítjuk a következő játékost

            this.CheckGame();
        }

        /// <summary>
        /// Játék betöltése.
        /// </summary>
        /// <param name="path">Elérési útvonal.</param>
        public void LoadGame(String path)
        {
            if (this.persistence == null)
                return;

            // végrehajtjuk a betöltést
            Player[] values = this.persistence.Load(path);

            if (values.Length != this.gameBoard.Length)
                throw new DataException("Error occured during game loading.");

            if (values.Count(value => value == Player.PlayerO) != values.Count(value => value == Player.PlayerX) && values.Count(value => value == Player.PlayerO) + 1 != values.Count(value => value == Player.PlayerX))
                // ellenőrizzük, hogy megfelelő számú lépés történt-e a játékosok részéről (leszámoljuk, hány lépés történt az egyes játékosok részéről)
                throw new DataException("Error occured during game loading.");

            this.StepNumber = values.Count(value => value != Player.NoPlayer); // leszámoljuk, hány lépés történt
            this.CurrentPlayer = this.StepNumber % 2 == 0 ? Player.PlayerX : Player.PlayerO; // a lépésszámból kijön az aktuális játékos

            // beállítjuk az értékeket
            for (Int32 i = 0; i < this.gameBoard.GetLength(0); i++)
                for (Int32 j = 0; j < this.gameBoard.GetLength(1); j++)
                {
                    this.gameBoard[i, j] = values[i * this.gameBoard.GetLength(0) + j];

                    this.OnFieldChanged(i, j, this.gameBoard[i, j]);
                }

            this.CheckGame();
        }

        /// <summary>
        /// Játék mentése.
        /// </summary>
        /// <param name="path">Elérési útvonal.</param>
        public void SaveGame(String path)
        {
            if (this.persistence == null)
                return;

            // az értékeket kimásoljuk egy új tömbbe
            Player[] values = new Player[this.gameBoard.Length];
            for (Int32 i = 0; i < this.gameBoard.GetLength(0); i++)
                for (Int32 j = 0; j < this.gameBoard.GetLength(1); j++)
                {
                    values[i * this.gameBoard.GetLength(0) + j] = this.gameBoard[i, j];
                }

            // végrehajtjuk a mentést
            this.persistence.Save(path, values);
        }

        #endregion

        #region Private methods

        /// <summary>
        /// Játék ellenőrzése.
        /// </summary>
        private void CheckGame() 
        {
            Player won = Player.NoPlayer;

            for (int i = 0; i < 3; ++i) // ellenőrzések végrehajtása
            {
                if (this.gameBoard[i, 0] != 0 && this.gameBoard[i, 0] == this.gameBoard[i, 1] && this.gameBoard[i, 1] == this.gameBoard[i, 2])
                    won = this.gameBoard[i, 0];
            }
            for (int i = 0; i < 3; ++i)
            {
                if (this.gameBoard[0, i] != 0 && this.gameBoard[0, i] == this.gameBoard[1, i] && this.gameBoard[1, i] == this.gameBoard[2, i])
                    won = this.gameBoard[0, i];
            }
            if (this.gameBoard[0, 0] != 0 && this.gameBoard[0, 0] == this.gameBoard[1, 1] && this.gameBoard[1, 1] == this.gameBoard[2, 2])
                won = this.gameBoard[0, 0];
            if (this.gameBoard[0, 2] != 0 && this.gameBoard[0, 2] == this.gameBoard[1, 1] && this.gameBoard[1, 1] == this.gameBoard[2, 0])
                won = this.gameBoard[0, 2];

            if (won != Player.NoPlayer) // ha valaki győzött
            {
                this.OnGameWon(won); // esemény kiváltása
            }
            else if (this.StepNumber == 9) // döntetlen játék
            {
                this.OnGameOver(); // esemény kiváltása
            }
        }

        #endregion

        #region Event triggers

        /// <summary>
        /// Játék megnyerésének eseménykiváltása.
        /// </summary>
        /// <param name="player">A győztes játékos.</param>
        private void OnGameWon(Player player)
        {
            if (this.GameWon != null)
                this.GameWon(this, new GameWonEventArgs(player));
        }

        /// <summary>
        /// Játék végének eseménykiváltása.
        /// </summary>
        private void OnGameOver()
        {
            if (this.GameOver != null)
                this.GameOver(this, EventArgs.Empty);
        }

        /// <summary>
        /// Mezőváltozás eseménykiváltása.
        /// </summary>
        /// <param name="x">Oszlop index.</param>
        /// <param name="y">Sor index.</param>
        /// <param name="player">Játékos.</param>
        private void OnFieldChanged(Int32 x, Int32 y, Player player)
        {
            if (this.FieldChanged != null)
                this.FieldChanged(this, new FieldChangedEventArgs(x, y, player));
        }

        #endregion
    }
}
